from collections import defaultdict


class FCFG_BASE:
    # 训练验证
    IS_MIXTURE_FIT = True  # 半精度训练 *
    FORWARD_COUNT = 1  # 多次前向一次反向 *
    IS_TRAIN = True  # 开启训练 *
    IS_VAL = False  # 开启验证 *
    IS_WARMUP = True  # 是否热身 *
    NUM_WARMUP = 2  # WARMUP 的轮数  n-1 *
    IS_EMA = False  # 参数平滑 *

    IS_TEST = True  # 开启测试 *
    IS_VISUAL = False  # 可视化模式 关联训练 和测试
    # MODE_VIS = 'bbox'  # 验证的 计算模式 'kps','bbox' 关系到 可视化 nms等 *
    NUM_VIS_Z = 3  # 测试时可视化的图片数 超过就不再可视 *

    IS_WRITER = False  # 使用CPU时有BUG 不能创建目录
    KEEP_WRITER = None  # 指定目录名称 FitExecutorBase 转换为 self.tb_writer

    IS_DEBUG = False  # 调试模式

    # 预测参数
    PRED_CONF_THR = 0.05  # 用于预测的阀值 *
    PRED_NMS_THR = 0.5  # 提高 conf 提高召回, 越小框越少 *
    PRED_SELECT_TOPK = 500  # 预测时topk *
    PRED_NMS_KEEP = 9999  # NMS每张图保留的个数 这个设置为100 有可能加速

    MAP_RECORD = defaultdict(lambda: [0.5, 0.5])
    MAP_RECORD.update({
        'type3': [0.762, 0.675],
        'type4': [0.732, 0.542],  # ema p542r368
        'face5': [0.988, 0.754],
        'face98': [0.470, 0.56],
        'widerface': [0.6, 0.5],
        'coco2017': [0.4, 0.4],
        'cocomin': [0.5, 0.5],
        'raccoon': [0.859, 0.672],
        'lsp': [0.859, 0.875],
    })
    MAPS_DEF_MAX = [0.5, 0.5]  # MAP的最大值 触发保存
    '''
    np.array([.26, .25, .25, .35, .35, 
                .79, .79, .72, .72, .62, .62, 
                1.07, 1.07, .87, .87, .89, .89])/10.0
    嘴眼 .25
    耳朵 .35
    肩 手关节 手腕 .79 .72 .62
    屁 脚关节 脚腕 1.07 .87 .89
    '''
    KPT_OKS_SIGMAS = None  # kps 验证必须 脸部25

    # 频率参数
    PRINT_FREQ = 1  # 打印频率 与 batch * PRINT_FREQ
    NUMS_TEST_DICT = {2: 1, 35: 1, 50: 1}  # 测试的起始和频率
    NUM_WEIGHT_SAVE_DICT = {2: 1, 35: 1, 50: 1}  # 保存频率

    # 数据集
    # BATCH_TRAIN = 3
    # BATCH_VAL = 1

    # 数据增强
    IS_MULTI_SCALE = False  # 开启多尺寸训练 只要用数据增强必须有
    MULTI_SCALE_VAL = [200, 300]  # 多尺寸训练的尺寸范围
    SIZE_WH_INPUT_TRAIN = (320, 320)
    SIZE_WH_INPUT_TEST = (320, 320)

    # MOSAIC 参数
    IS_MOSAIC = False
    IS_MIXUP = False
    IS_MOSAIC_FILTER = False
    THR_WH = 5
    THR_AREA = 25

    # 其它动态参数
    PATH_SAVE_WEIGHT = ''  # ***
    SAVE_WEIGHT_NAME = ''  # ***
    PATH_HOST = 'nan'  # 用于多环境切换主机
    NUM_WORKERS = 4
    FUN_LOCK = None  # 冻结训练参数
    NUMS_LOCK_WEIGHT_DICT = None  # np.array([1, 5, 12]) 锁定权重分段训练  [1, 10] epoch 与 FUN_LOCK 匹配 必须由小到大
    LR_BASE = 1e-3
    START_EPOCH = 1
    END_EPOCH = 300
    TRANSFORM_TRAIN = None
    TRANSFORM_TEST = None
    DATA_INFO = None  # 这个用于存一些data信息 DATA_INFO对象
    IS_JUMP = False  # 当coco数据集有没有目标的GT时 默认可参与训练 不跳过  boxes为空需要处理

    # 可调参数 这个要在 fun_get_data 的后面

    def __str__(self) -> str:
        s = '---------------- CFG参数 ------------------\n'
        for name in dir(self):  # 显示对象的所有属性
            if not name.startswith('__'):  # 过滤掉系统自带的
                s += ('\t' + name + ' : ' + str(getattr(self, name)) + '\n')
        return s


if __name__ == '__main__':
    print(str(FCFG_BASE()))



import random
import numpy as np
import cv2
import math
import matplotlib.pyplot as plt


def _flip_matrix4nanodet(prob=0.5):
    '''
    随机水平翻转
    :param prob: 概率
    :return:
    '''
    F = np.eye(3)
    if random.random() < prob:
        F[0, 0] = -1
    return F


def _perspective_matrix4nanodet(perspective=0):
    """

    :param perspective:
    :return:
    """
    P = np.eye(3)
    P[2, 0] = random.uniform(-perspective, perspective)  # x perspective (about y)
    P[2, 1] = random.uniform(-perspective, perspective)  # y perspective (about x)
    return P


def _rotation_matrix4nanodet(degree=0):
    """

    :param degree:
    :return:
    """
    R = np.eye(3)
    a = random.uniform(-degree, degree)
    R[:2] = cv2.getRotationMatrix2D(angle=a, center=(0, 0), scale=1)
    return R


def _scale_matrix4nanodet(ratio=(1, 1)):
    """

    :param width_ratio:
    :param height_ratio:
    """
    Scl = np.eye(3)
    scale = random.uniform(*ratio)
    Scl[0, 0] *= scale
    Scl[1, 1] *= scale
    return Scl


def _stretch_matrix4nanodet(width_ratio=(1, 1), height_ratio=(1, 1)):
    """

    :param width_ratio:
    :param height_ratio:
    """
    Str = np.eye(3)
    Str[0, 0] *= random.uniform(*width_ratio)
    Str[1, 1] *= random.uniform(*height_ratio)
    return Str


def _shear_matrix4nanodet(degree):
    """

    :param degree:
    :return:
    """
    Sh = np.eye(3)
    Sh[0, 1] = math.tan(random.uniform(-degree, degree) * math.pi / 180)  # x shear (deg)
    Sh[1, 0] = math.tan(random.uniform(-degree, degree) * math.pi / 180)  # y shear (deg)
    return Sh


def _translate_matrix4nanodet(translate, width, height):
    """

    :param translate:
    :return:
    """
    T = np.eye(3)
    T[0, 2] = random.uniform(0.5 - translate, 0.5 + translate) * width  # x translation
    T[1, 2] = random.uniform(0.5 - translate, 0.5 + translate) * height  # y translation
    return T


def _resize_matrix4nanodet(raw_shape, dst_shape, keep_ratio):
    """
    Get resize matrix for resizing raw img to input size
    :param raw_shape: (width, height) of raw image
    :param dst_shape: (width, height) of input image
    :param keep_ratio: whether keep original ratio
    :return: 3x3 Matrix
    """
    r_w, r_h = raw_shape
    d_w, d_h = dst_shape
    Rs = np.eye(3)
    if keep_ratio:
        C = np.eye(3)
        C[0, 2] = - r_w / 2
        C[1, 2] = - r_h / 2

        if r_w / r_h < d_w / d_h:
            ratio = d_h / r_h
        else:
            ratio = d_w / r_w
        Rs[0, 0] *= ratio
        Rs[1, 1] *= ratio

        T = np.eye(3)
        T[0, 2] = 0.5 * d_w
        T[1, 2] = 0.5 * d_h
        return T @ Rs @ C
    else:
        Rs[0, 0] *= d_w / r_w
        Rs[1, 1] *= d_h / r_h
        return Rs


def _warp_boxes4nanodet(boxes, M, width, height):
    n = len(boxes)
    if n:
        # warp points
        xy = np.ones((n * 4, 3))
        xy[:, :2] = boxes[:, [0, 1, 2, 3, 0, 3, 2, 1]].reshape(n * 4, 2)  # x1y1, x2y2, x1y2, x2y1
        xy = xy @ M.T  # transform
        xy = (xy[:, :2] / xy[:, 2:3]).reshape(n, 8)  # rescale
        # create new boxes
        x = xy[:, [0, 2, 4, 6]]
        y = xy[:, [1, 3, 5, 7]]
        xy = np.concatenate((x.min(1), y.min(1), x.max(1), y.max(1))).reshape(4, n).T
        # clip boxes
        xy[:, [0, 2]] = xy[:, [0, 2]].clip(0, width)
        xy[:, [1, 3]] = xy[:, [1, 3]].clip(0, height)
        return xy.astype(np.float32)
    else:
        return boxes


def data_heighten4nanodet(img_np, target, dst_shape, keep_ratio=True):
    h_src, w_src = img_np.shape[:2]  # shape(h,w,c)

    # center
    C = np.eye(3)
    C[0, 2] = - w_src / 2
    C[1, 2] = - h_src / 2

    # do not change the order of mat mul
    if random.randint(0, 1):
        # 随机透视0.0
        P = _perspective_matrix4nanodet()
        C = P @ C
    if random.uniform(0, 1) > 0.5:
        # 缩放[0.6, 1.4]
        Scl = _scale_matrix4nanodet([0.6, 1.4])
        C = Scl @ C
    if random.randint(0, 1):
        # 拉伸[[1, 1], [1, 1]]
        Str = _stretch_matrix4nanodet(width_ratio=(1, 1), height_ratio=(1, 1))
        C = Str @ C
    if random.randint(0, 1):
        # 旋转没用
        R = _rotation_matrix4nanodet(degree=0)
        C = R @ C
    if random.randint(0, 1):
        # 切变没用
        Sh = _shear_matrix4nanodet(degree=0)
        C = Sh @ C
    if random.randint(0, 1):
        # 翻转0.5
        F = _flip_matrix4nanodet(prob=1.0)
        C = F @ C
    if random.randint(0, 1):
        # 随机移动
        T = _translate_matrix4nanodet(0.5, w_src, h_src)
    else:
        T = _translate_matrix4nanodet(0, w_src, h_src)
    M = T @ C
    # M = T @ Sh @ R @ Str @ P @ C
    ResizeM = _resize_matrix4nanodet((w_src, h_src), dst_shape, keep_ratio)
    M = ResizeM @ M
    img_np_new = cv2.warpPerspective(img_np, M, dsize=tuple(dst_shape))
    target['warp_matrix'] = M
    if 'boxes' in target:
        boxes = target['boxes']
        target['boxes'] = _warp_boxes4nanodet(boxes, M, dst_shape[0], dst_shape[1])
    if 'gt_masks' in target:
        for i, mask in enumerate(target['gt_masks']):
            target['gt_masks'][i] = cv2.warpPerspective(mask, M, dsize=tuple(dst_shape))

    return img_np_new


def filter_box4yolox(box1, box2, wh_thr=2, ar_thr=20, area_thr=0.2):
    # box1(4,n), box2(4,n)
    # Compute candidate boxes which include follwing 5 things:
    # box1 before augment, box2 after augment, wh_thr (pixels), aspect_ratio_thr, area_ratio
    w1, h1 = box1[2] - box1[0], box1[3] - box1[1]
    w2, h2 = box2[2] - box2[0], box2[3] - box2[1]
    ar = np.maximum(w2 / (h2 + 1e-16), h2 / (w2 + 1e-16))  # aspect ratio
    return (
            (w2 > wh_thr)
            & (h2 > wh_thr)
            & (w2 * h2 / (w1 * h1 + 1e-16) > area_thr)
            & (ar < ar_thr)
    )  # candidates


def data_heighten4yolox(img, target, degrees=10, translate=0.1,
                        scale=(0.5, 1.5), shear=10, perspective=0.0,
                        border=(0, 0),
                        ):
    if 'boxes' in target:
        g_ltrb_mosaic = target['boxes']

        height = img.shape[0] + border[0] * 2  # shape(h,w,c)
        width = img.shape[1] + border[1] * 2

        # Center
        C = np.eye(3)
        C[0, 2] = -img.shape[1] / 2  # x translation (pixels)
        C[1, 2] = -img.shape[0] / 2  # y translation (pixels)

        # Rotation and Scale 旋转及缩放
        R = np.eye(3)
        a = random.uniform(-degrees, degrees)
        # a += random.choice([-180, -90, 0, 90])  # add 90deg rotations to small rotations
        s = random.uniform(scale[0], scale[1])
        # s = 2 ** random.uniform(-scale, scale)
        R[:2] = cv2.getRotationMatrix2D(angle=a, center=(0, 0), scale=s)

        # Shear 剪切
        S = np.eye(3)
        S[0, 1] = math.tan(random.uniform(-shear, shear) * math.pi / 180)  # x shear (deg)
        S[1, 0] = math.tan(random.uniform(-shear, shear) * math.pi / 180)  # y shear (deg)

        # Translation 平移
        T = np.eye(3)
        T[0, 2] = (random.uniform(0.5 - translate, 0.5 + translate) * width)  # x translation (pixels)
        T[1, 2] = (random.uniform(0.5 - translate, 0.5 + translate) * height)  # y translation (pixels)

        # Combined rotation matrix
        M = T @ S @ R @ C  # order of operations (right to left) is IMPORTANT

        ###########################
        # For Aug out of Mosaic
        # s = 1.
        # M = np.eye(3)
        ###########################

        if (border[0] != 0) or (border[1] != 0) or (M != np.eye(3)).any():  # image changed
            if perspective:
                img = cv2.warpPerspective(img, M, dsize=(width, height), borderValue=(128, 128, 128))
            else:  # affine
                img = cv2.warpAffine(img, M[:2], dsize=(width, height), borderValue=(128, 128, 128))

        # Transform label coordinates
        n = len(g_ltrb_mosaic)
        if n:
            # warp points
            xy = np.ones((n * 4, 3))
            xy[:, :2] = g_ltrb_mosaic[:, [0, 1, 2, 3, 0, 3, 2, 1]].reshape(n * 4, 2)  # x1y1, x2y2, x1y2, x2y1
            xy = xy @ M.T  # transform
            if perspective:
                xy = (xy[:, :2] / xy[:, 2:3]).reshape(n, 8)  # rescale
            else:  # affine
                xy = xy[:, :2].reshape(n, 8)

            # create new boxes
            x = xy[:, [0, 2, 4, 6]]
            y = xy[:, [1, 3, 5, 7]]
            xy = np.concatenate((x.min(1), y.min(1), x.max(1), y.max(1))).reshape(4, n).T

            # clip boxes
            xy[:, [0, 2]] = xy[:, [0, 2]].clip(0, width)
            xy[:, [1, 3]] = xy[:, [1, 3]].clip(0, height)

            # filter candidates
            i = filter_box4yolox(box1=g_ltrb_mosaic[:, :4].T * s, box2=xy.T)
            # g_ltrb_mosaic = g_ltrb_mosaic[i]
            # g_ltrb_mosaic[:, :4] = xy[i]
            target['boxes'] = xy[i]
            target['labels'] = target['labels'][i]
    else:
        raise Exception('boxes 为空')
    return img, target


if __name__ == '__main__':
    file_img = r'/_test_pic/test_99.jpg'  # 650 466
    img_np = cv2.imread(file_img)  # 这个打开是hwc bgr
    img_np = cv2.cvtColor(img_np, cv2.COLOR_BGR2RGB)
    target = {}
    img_np_new = data_heighten4nanodet(img_np, target, dst_shape=(320, 320))
    plt.imshow(img_np_new)
    plt.show()
    # img_np = cv2.cvtColor(img_np, cv2.COLOR_BGR2RGB)
    pass



import random

import cv2
import torch
import numpy as np

from ftools.ai.object_detection.boxes.f_boxes import adjust_box_anns, fverify_bbox
from ftools.ai.datas.dta_heighten.f_data_heighten_matrix import filter_box4yolox
from ftools.ai.datas.dta_heighten.f_data_heighten_matrix import data_heighten4yolox


def get_mosaic_coordinate(mosaic_index, xc, yc, w_img, h_img, input_h, input_w):
    '''
    以 xc, yc 为固定点
    :param mosaic_index:
    :param xc:
    :param yc:
    :param w_img:
    :param h_img:
    :param input_h:
    :param input_w:
    :return:
        背景的 ltrb
        粘图的 ltrb
    '''
    # index0 to top left part of image
    if mosaic_index == 0:
        x1, y1, x2, y2 = max(xc - w_img, 0), max(yc - h_img, 0), xc, yc
        ltrb_pic = w_img - (x2 - x1), h_img - (y2 - y1), w_img, h_img
    # index1 to top right part of image
    elif mosaic_index == 1:
        x1, y1, x2, y2 = xc, max(yc - h_img, 0), min(xc + w_img, input_w * 2), yc
        ltrb_pic = 0, h_img - (y2 - y1), min(w_img, x2 - x1), h_img
    # index2 to bottom left part of image
    elif mosaic_index == 2:
        x1, y1, x2, y2 = max(xc - w_img, 0), yc, xc, min(input_h * 2, yc + h_img)
        ltrb_pic = w_img - (x2 - x1), 0, w_img, min(y2 - y1, h_img)
    # index2 to bottom right part of image
    elif mosaic_index == 3:
        x1, y1, x2, y2 = xc, yc, min(xc + w_img, input_w * 2), min(input_h * 2, yc + h_img)  # noqa
        ltrb_pic = 0, 0, min(w_img, x2 - x1), min(y2 - y1, h_img)
    # ltrb_mosaic, ltrb_pic
    else:
        raise Exception('mosaic_index = %s 越界 ' % mosaic_index)
    return (x1, y1, x2, y2), ltrb_pic


def get_mosaic_img(imgs, targets, size_hw_m=(640, 640),
                   thr_wh=5, thr_area=25, is_filter=False):
    '''

    :param id_img:
    :param size_hw_m: 输入 mosaic 的图片尺寸 mosaic尺寸为  *2
    :param is_filter: 启用框过滤过滤   thr_wh=5, thr_area=25
    :return:
    '''

    # yc, xc = s, s  # mosaic center x, y
    yc = int(random.uniform(0.5 * size_hw_m[0], 1.5 * size_hw_m[0]))
    xc = int(random.uniform(0.5 * size_hw_m[1], 1.5 * size_hw_m[1]))

    target_mosaic = {}
    size_mosaic = (size_hw_m[0] * 2, size_hw_m[1] * 2)
    # target_mosaic['image_id'] = np.array([]) # 这个训练没用
    target_mosaic['size'] = np.array(size_mosaic)
    target_mosaic['boxes'] = np.empty((0, 4))  # 创建空数组
    target_mosaic['labels'] = np.empty((0))

    for i_mosaic, (img_np, target), in enumerate(zip(imgs, targets)):
        h_img, w_img, = img_np.shape[:2]
        scale = min(1. * size_hw_m[0] / h_img, 1. * size_hw_m[1] / w_img)
        # resize 的参数是 w,h
        img_np = cv2.resize(img_np, dsize=(int(w_img * scale), int(h_img * scale)), interpolation=cv2.INTER_LINEAR)
        h_img, w_img, c = img_np.shape  # resize后的尺寸
        if i_mosaic == 0:
            img_mosaic = np.full((*size_mosaic, c), 114, dtype=np.uint8)
        ltrb_mosaic, ltrb_pic = get_mosaic_coordinate(mosaic_index=i_mosaic, xc=xc, yc=yc,
                                                      w_img=w_img, h_img=h_img,
                                                      input_h=size_hw_m[0], input_w=size_hw_m[1], )
        (l_x1, l_y1, l_x2, l_y2), (s_x1, s_y1, s_x2, s_y2) = ltrb_mosaic, ltrb_pic
        img_mosaic[l_y1:l_y2, l_x1:l_x2] = img_np[s_y1:s_y2, s_x1:s_x2]
        offw, offh = l_x1 - s_x1, l_y1 - s_y1

        # g_ltrb = target['boxes']  # 前面已处理了有效性
        if len(target['boxes']) > 0:
            target['boxes'][:, 0] = scale * target['boxes'][:, 0] + offw
            target['boxes'][:, 1] = scale * target['boxes'][:, 1] + offh
            target['boxes'][:, 2] = scale * target['boxes'][:, 2] + offw
            target['boxes'][:, 3] = scale * target['boxes'][:, 3] + offh

        target_mosaic['boxes'] = np.concatenate([target_mosaic['boxes'], target['boxes']], 0)
        target_mosaic['labels'] = np.concatenate([target_mosaic['labels'], target['labels']], 0)

    # 图内剪切
    np.clip(target_mosaic['boxes'][:, ::2], 0, 2 * size_hw_m[1], out=target_mosaic['boxes'][:, ::2])
    np.clip(target_mosaic['boxes'][:, 1::2], 0, 2 * size_hw_m[0], out=target_mosaic['boxes'][:, 1::2])

    if is_filter:
        mask = fverify_bbox(target_mosaic['boxes'], thr_wh=thr_wh, thr_area=thr_area)
        target_mosaic['boxes'] = target_mosaic['boxes'][mask]
        target_mosaic['labels'] = target_mosaic['labels'][mask]

    # 原装数据增加过滤
    img_mosaic, target_mosaic = data_heighten4yolox(img_mosaic, target_mosaic)

    return img_mosaic, target_mosaic


def mixup(img_src, target_src, img_det, target_det, size_hw_m, mixup_scale=(0.5, 1.5)):
    if isinstance(target_det['boxes'], np.ndarray):
        fun_copy = np.copy
        fun_flip = np.flip
        fun_cat = np.concatenate
    elif isinstance(target_det['boxes'], torch.Tensor):
        fun_copy = torch.clone
        fun_flip = torch.flip
        fun_cat = torch.cat
    else:
        raise Exception('输入类型错误 %s' % type(target_det['boxes']))

    jit_factor = random.uniform(*mixup_scale)
    FLIP = random.uniform(0, 1) > 0.5
    # FLIP = True

    # (640,640) 背景
    img_temp = np.ones((size_hw_m[0], size_hw_m[1], 3)) * 114.0
    h_img, w_img, = img_det.shape[:2]
    scale_ratio_det = min(1. * size_hw_m[0] / h_img, 1. * size_hw_m[1] / w_img)
    # 缩小到背景中
    resized_img = cv2.resize(img_det,
                             (int(w_img * scale_ratio_det), int(h_img * scale_ratio_det)),
                             interpolation=cv2.INTER_LINEAR)
    # 把图装到
    img_temp[: int(h_img * scale_ratio_det), : int(w_img * scale_ratio_det)] = resized_img
    # 随机缩放
    img_temp = cv2.resize(img_temp,
                          (int(img_temp.shape[1] * jit_factor), int(img_temp.shape[0] * jit_factor)),
                          )
    scale_ratio_det *= jit_factor
    # f_show_od_np4plt_v3(img_temp, )
    if FLIP:
        img_temp = img_temp[:, ::-1, :]  # 水平翻转
        # f_show_od_np4plt_v3(img_temp)
        pass
    # 这里有可能放大或缩小
    origin_h, origin_w = img_temp.shape[:2]
    target_h, target_w = img_src.shape[:2]

    img_pad = np.zeros((max(origin_h, target_h), max(origin_w, target_w), 3)).astype(np.uint8)
    # 这里有是剪切或填充
    img_pad[:origin_h, :origin_w] = img_temp

    x_offset, y_offset = 0, 0  # 如果是填充 随机确定一个左上角点
    if img_pad.shape[0] > target_h:
        y_offset = random.randint(0, img_pad.shape[0] - target_h - 1)
    if img_pad.shape[1] > target_w:
        x_offset = random.randint(0, img_pad.shape[1] - target_w - 1)
    padded_cropped_img = img_pad[y_offset: y_offset + target_h, x_offset: x_offset + target_w]

    # 处理bbox
    boxes_det_origin = adjust_box_anns(target_det['boxes'],
                                       scale_ratio_det, 0, 0,
                                       origin_w, origin_h
                                       )
    if FLIP:
        # torch 不支持 ::-1
        # boxes_det_origin[:, 0::2] = (origin_w - boxes_det_origin[:, 0::2][:, ::-1])
        # 这里反转的水平
        boxes_det_origin[:, 0::2] = (origin_w - fun_flip(boxes_det_origin[:, 0::2], 1))

    # cp_bboxes_transformed_np = cp_bboxes_origin_np.copy()
    # boxes_det_origin_transformed = boxes_det_origin.clone()
    boxes_det_origin_transformed = fun_copy(boxes_det_origin)
    boxes_det_origin_transformed[:, 0::2] = np.clip(boxes_det_origin_transformed[:, 0::2] - x_offset,
                                                    0, target_w
                                                    )
    boxes_det_origin_transformed[:, 1::2] = np.clip(boxes_det_origin_transformed[:, 1::2] - y_offset,
                                                    0, target_h
                                                    )
    keep_list = filter_box4yolox(boxes_det_origin.T, boxes_det_origin_transformed.T, 5)

    if keep_list.sum() >= 1.0:
        # f_show_od_np4plt_v3(padded_cropped_img, cp_bboxes_transformed_np[keep_list])
        # 这里合并
        target_src['labels'] = fun_cat([target_src['labels'], target_det['labels'][keep_list]], 0)
        target_src['boxes'] = fun_cat([target_src['boxes'], boxes_det_origin_transformed[keep_list]], 0)
        img_src = img_src.astype(np.float32)
        img_src = 0.5 * img_src + 0.5 * padded_cropped_img.astype(np.float32)

    return img_src.astype(np.uint8), target_src


if __name__ == '__main__':
    from ftools.ai.datas.z_dataloader import get_data_cocomin, get_data_type3
    from ftools.ai.picture.f_show import f_show_od_np4plt_v3
    from ftools.ai.CONFIG_BASE import FCFG_BASE

    path_host = 'M:'
    mode = 'bbox'
    cfg = FCFG_BASE()
    cfg.IS_TRAIN = True
    cfg.IS_VAL = False
    cfg.IS_TEST = False
    cfg.IS_MULTI_SCALE = False
    cfg.num_workers = 0
    cfg.SIZE_WH_INPUT_TRAIN = (640, 640)
    cfg.is_mosaic = False
    data_info, dataloader_train, dataloader_test = get_data_type3(
        cfg,
        path_host=path_host,
        mode_train=mode, batch_train=10
    )

    dataset = dataloader_train.dataset

    for i in range(len(dataset)):
        id_img = dataset.ids_img[i]  # 公用
        ids_img = [id_img] + np.random.choice(dataset.ids_img, 3).tolist()
        imgs, targets = [], []
        for id_img in ids_img:
            img_np, target = dataset.fgetitem(id_img)
            imgs.append(img_np)
            targets.append(target)

        img_np, target = get_mosaic_img(imgs=imgs,
                                        targets=targets,
                                        size_hw_m=cfg.SIZE_WH_INPUT_TRAIN)

        # img_src, target_src = dataset.fgetitem(id_img)
        # _id_img = np.random.choice(dataset.ids_img, 1)[0]
        # img_det, target_det = dataset.fgetitem(_id_img)  # dataset 已确保有目标
        # img_np, target = mixup(img_src, target_src, img_det, target_det, cfg.SIZE_WH_INPUT_TRAIN )
        # # f_show_od_np4plt_v3(img_np, target['boxes'], g_texts=target['labels'])

        # 数据增强
        # img_mosaic, g_ltrb_mosaic = data_heighten4yolox(
        #     img_mosaic, g_ltrb_mosaic,
        #     degrees=10,
        #     translate=0.1,
        #     scale=(0.5, 1.5),
        #     shear=2.,
        #     perspective=0.,
        #     border=[-size_hw_m[0] // 2, -size_hw_m[1] // 2],
        # )

        text = [str(l) for l in target['labels']]
        f_show_od_np4plt_v3(img_np, target['boxes'], g_texts=text, )

        pass
